#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-07-06
# File  : vsphere_topo/lib/utils.py

from collections.abc import Mapping, MutableMapping, MutableSequence, Sequence
from json import dumps as json_dunps, loads as json_loads
from os import environ
from pathlib import Path
from typing import Any, Final, Tuple

from livestatus import MultiSiteConnection, SiteConfigurations, SiteId

EMBLEM_CLUSTER: Final[str] = 'icon_plugins_hw'
EMBLEM_DATA_CENTER: Final[str] = 'icon_cloud'
EMBLEM_DATA_STORE: Final[str] = 'icon_services_green'

ICON_VCENTER: Final[str] = 'icon_topic_hosts'

PARAM_ADD_DUMMY_TOPOLOGIES: Final[str] = 'add_dummy_topologies'
PARAM_CHANGE_CASE: Final[str] = 'change_case'
PARAM_CLUSTER: Final[str] = 'cluster'
PARAM_CUSTOM_EMBLEM: Final[str] = 'custom_emblem'
PARAM_DATA_CENTER: Final[str] = 'data_center'
PARAM_DATA_STORE: Final[str] = 'data_store'
PARAM_DATA_STORE_AS_SERVICE: Final[str] = 'data_sore_as_service'
PARAM_DEFAULT_EMBLEM: Final[str] = 'default_emblem'
PARAM_DONT_ADD_VC_AS_VM: Final[str] = 'dont_add_vc_as_vm'
PARAM_HOST_SYSTEMS: Final[str] = 'host_systems'
PARAM_KEEP_DOMAIN: Final[str] = 'keep_domain'
PARAM_LOWER: Final[str] = 'lower'
PARAM_MAKE_DEFAULT: Final[str] = 'make_default'
PARAM_NO_CHANGE: Final[str] = 'no_change'
PARAM_NO_EMBLEM: Final[str] = 'no_emblem'
PARAM_PREFIX: Final[str] = 'prefix'
PARAM_UPPER: Final[str] = 'upper'
PARAM_VCENTER: Final[str] = 'vcenter_icon'
PARAM_VM_NAMES: Final[str] = 'vm_names'

PICTURE_TYPE_EMBLEM: Final[str] = 'emblem'
PICTURE_TYPE_ICON: Final[str] = 'icon'

RULE_SET_NAME_VSPHERE_TOPO: Final[str] = 'vsphere_topo'

OMD_ROOT = environ["OMD_ROOT"]
BASE_TOPO_PATH: Final[str] = f'{OMD_ROOT}/var/check_mk/topology/data'


def adjust_name(name: str, params: Mapping[str, Any]) -> str:
    class Case:
        upper: str = PARAM_UPPER
        lower: str = PARAM_LOWER
        no_change: str = PARAM_NO_CHANGE

    if params.get(PARAM_KEEP_DOMAIN) is None:
        name = name.split('.')[0]

    match params.get(PARAM_CHANGE_CASE, PARAM_NO_CHANGE):
        case Case.upper:
            name = name.upper()
        case Case.lower:
            name = name.lower()
        case Case.no_change | _:  # all else
            pass

    if (prefix := params.get(PARAM_PREFIX)) is not None:
        name = f'{prefix.strip()}{name.strip()}'

    return name.strip()


def get_emblem(emblem: str, params: Tuple[str, bool | str] | None = None) -> str | None:
    class Emblem:
        no_emblem: str = PARAM_NO_EMBLEM
        default_emblem: str = PARAM_DEFAULT_EMBLEM
        custom_emblem: str = PARAM_CUSTOM_EMBLEM

    match params:
        case None:
            return emblem
        case (Emblem.no_emblem, True):
            return None
        case (Emblem.default_emblem, True):
            return emblem
        case _:  # custom_emblem
            return params[1].strip()


def save_topology(data: Mapping, sub_directory: str) -> None:
    """
    Save the topology as json file under $OMD_ROOT/var/check_mk/topology/data/{sub_directory}.data_{data['name']}.json
    the filename will be changed to lower case.
    Args:
        data: the topology data
        sub_directory: the subdirectory were to save the data under ~/var/check_mk/topology/data/

    Returns:
        None
    """
    file_name = f'data_{data["name"]}.json'.lower()
    save_file = Path(f'{BASE_TOPO_PATH}/{sub_directory}/{file_name}')
    save_file.parent.mkdir(exist_ok=True, parents=True)
    save_file.write_text(json_dunps(data))


def make_topo_default(sub_directory: str, make_default: bool) -> None:
    """
    Create the symlink "default" to $OMD_ROOT/var/check_mk/topology/data/{sub_directory} in
    $OMD_ROOT/var/check_mk/topology/data/ if it don't exist or mage_default is True
    Args:
        sub_directory: the subdirectory under ~/var/check_mk/topology/data/ thaht become default
        make_default: if True, create the symlink "default" with path as target

    Returns:
        None
    """

    target_path = f'{BASE_TOPO_PATH}/{sub_directory}'

    if not Path(f'{BASE_TOPO_PATH}/default').exists():
        make_default = True
    if make_default:
        Path(f'{BASE_TOPO_PATH}/default').unlink(missing_ok=True)
        Path(f'{BASE_TOPO_PATH}/default').symlink_to(target=Path(target_path), target_is_directory=True)


def get_topologies() -> Sequence[str | None]:
    """
    Returns a list of topology names form the default typology directory.

    Returns:
        List of str ie: ['CDP', 'LLDP']
    """
    path: str = f'{BASE_TOPO_PATH}/default'
    if not Path(path).exists():
        return []

    files = [f for f in Path(path).glob('*.json') if f.is_file()]
    return [
        json_loads(Path(file).read_text())['name'] for file in files if
        json_loads(Path(file).read_text()).get('name') is not None
    ]


def add_dummy_topologies(sub_directory: str):
    path: str = f'{BASE_TOPO_PATH}/default'
    if Path(path).exists() and not Path(f'{path}/data_vsphere.json').exists():  # don't overwrite existing topology
        dummy_topology = {'version': 1, 'name': 'vSphere', 'objects': {}, 'connections': []}
        save_topology(
            data=dummy_topology,
            sub_directory='default',
        )

    for topology in get_topologies():
        if not Path(f'{BASE_TOPO_PATH}/{sub_directory}/data_{topology.lower()}.json').exists():
            save_topology(
                data={'version': 1, 'name': topology, 'objects': {}, 'connections': []},
                sub_directory=sub_directory
            )


#
#  live status
#
class LiveStatusConnection(object):
    def __init__(self):
        self.sites: SiteConfigurations = SiteConfigurations({})
        self.sites_mk = Path(f'{OMD_ROOT}/etc/check_mk/multisite.d/sites.mk')
        self.socket_path = f'unix:{OMD_ROOT}/tmp/run/live'
        if self.sites_mk.exists():
            # make eval() "secure"
            # https://realpython.com/python-eval-function/#minimizing-the-security-issues-of-eval
            _code = compile(self.sites_mk.read_text(), "<string>", "eval")
            allowed_names = ['sites', 'update']
            for name in _code.co_names:
                if name not in allowed_names:
                    raise NameError(f'Use of {name} in {self.sites_mk.name} not allowed.')

            sites_raw: MutableMapping = {}
            eval(self.sites_mk.read_text(), {"__builtins__": {}}, {"sites": sites_raw})

            for site, data in sites_raw.items():
                self.sites.update({site: {
                    'alias': data['alias'],
                    'timeout': data['timeout'],
                }})
                if data['socket'] == ('local', None):
                    self.sites[site]['socket'] = self.socket_path
                else:
                    protocol, socket = data['socket']
                    address, port = socket['address']
                    self.sites[site]['socket'] = f'{protocol}:{address}:{port}'
                    self.sites[site]['tls'] = socket['tls']
        else:
            self.sites.update({SiteId('local'): {
                'alias': 'Local site',
                'timeout': 5,
                'socket': self.socket_path
            }})

        self.c = MultiSiteConnection(self.sites)
        dead_sites = [site['site']['alias'] for site in self.c.dead_sites().values()]
        if dead_sites:
            self.c.set_only_sites(self.c.alive_sites())

    def query(self, query: str):
        data: MutableSequence[Tuple[str, str]] = self.c.query(query=query)

        if data:
            return data
        else:
            return None
